/*## (c) SECURITY EXPLORATIONS    2011 poland                                #*/
/*##     http://www.security-explorations.com                                #*/

/* RESEARCH MATERIAL:	SRP-2018-01                                           */
/* SlimCORE disassembler                                                      */

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */

/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

import java.lang.*;
import java.io.*;
import java.util.*;
import java.security.*;

public class SCDisasm {
 public static final String STTKDMA_DRIVER = "sttkdma_core_user.ko";

 public static final int MAGIC1_CODE_START = 0x52455300; //"RES" string
 public static final int MAGIC2_CODE_START = 0x00002000; //add r0,r0,r0,#0000

 public static final int MAGIC1_CODE_END   = 0x000d8400; //jmp r13
 public static final int MAGIC2_CODE_END   = 0x00000000;

 public static final int MAGIC1_VER        = 0x5354544b; //"STTK"
 public static final int MAGIC2_VER        = 0x444d412d; //"DMA-"

 public static final int DATA_SIZE   = 0x0484;

 public static final int DSTART      = 0x4000;
 public static final int IOSTART     = 0x5e00;

 public static final int DISASM          = 0x01;
 public static final int EXTRACT_CODE    = 0x02;
 public static final int EXTRACT_DATA    = 0x03;
 public static final int STATS_PRINT_UNK = 0x04;
 public static final int STATS_PRINT_ALL = 0x05;

 public static final int DUMP_FROM_DRIVER = 0x01;
 public static final int DUMP_FROM_FILES  = 0x02;

 public static final int ANNOTATION_PADDING = 70;

 public static String drv_name;
 public static String code_name;
 public static String data_name;
 public static String ann_name;

 private static int drv[];

 private static int code[];
 private static int csize;
 private static int data[];
 private static int dsize;

 private static boolean bblocks[];

 private static boolean fulldisasm;
 private static int access_method;

 private static Hashtable syms;

 private static Hashtable annotations;
 private static Hashtable comments;

 public static void add_sym(int val,String name) {
  syms.put(new Integer(val),name);
 }

 public static String lookup_sym(int val) {
  return (String)syms.get(new Integer(val));
 }

 public static void init_syms() {
  try {
   syms=new Hashtable();

   add_sym(0x4004,"DMA CONFIG");
   add_sym(0x4008,"DMA src");
   add_sym(0x400c,"DMA dest");
   add_sym(0x4010,"DMA size");
   add_sym(0x4018,"TK CONFIG");
   add_sym(0x401c,"TK CMD");
   add_sym(0x4020,"TK DATA[0]");
   add_sym(0x4024,"TK DATA[4]");
   add_sym(0x4028,"TK DATA[8]");
   add_sym(0x402c,"TK DATA[c]");
   add_sym(0x4030,"CUSTOMER MODE");

   add_sym(0x5e2c,"PUBLIC_ID");
  } catch(Throwable t) {}
 }

 public static String hex_value(int val,int max) {
  String s="";
  for(int i=0;i<max;i++) {
   s+="0";
  }

  s+=Integer.toHexString(val);

  return s.substring(s.length()-max);
 }

 public static String lookup_version() {
  String ver=null;

  for(int i=0;i<drv.length-2*4;i+=4) {
   int w1=read_dword(drv,i);
   int w2=read_dword(drv,i+4);

   if ((w1==MAGIC1_VER)&&(w2==MAGIC2_VER)) {
    ByteArrayOutputStream baos=new ByteArrayOutputStream();
    int pos=i;

    while(drv[pos]!=0) {
     int c=drv[pos++];
     baos.write(c);               
    }

    ver=new String(baos.toByteArray());
   }
  }

  return ver;
 }

 public static String sha1(int data[]) throws Exception {
  byte buf[]=new byte[data.length];

  for(int i=0;i<data.length;i++) {
   buf[i]=(byte)data[i];
  }

  MessageDigest md=MessageDigest.getInstance("SHA-1");
  md.update(buf);

  byte sha1[]=md.digest();

  String res="";

  for(int i=0;i<sha1.length;i++) {
   res+=hex_value(((int)sha1[i])&0xff,2);
  }

  return res;
 }

 public static int opcode(int off) {
  int val=0;
  
  val=code[off];
  val|=code[off+1]<<8;
  val|=code[off+2]<<16;
  val|=code[off+3]<<24;

  return val;  
 }

 public static int data(int off) {
  int val=0;
  
  val=data[off];
  val|=data[off+1]<<8;
  val|=data[off+2]<<16;
  val|=data[off+3]<<24;

  return val;  
 }

 public static void mark_target(int idx) {
  if (fulldisasm) {
   bblocks[idx]=true;
  }
 }

 private static String do_pad(int cnt) {
  String s="";

  for(int i=0;i<cnt;i++) {
   s+=" ";
  }

  return s;
 }

 private static void dump_code() {
  System.out.println("- disassembling");

  System.out.println("[CODE]");

  for(int i=0;i<csize;i+=4) {
   int opcode=opcode(i);
   disasm(opcode);
  }

  for(int i=0;i<csize;i+=4) {
   int opcode=opcode(i);

   String pfx="  ";

   if (bblocks[i>>2]) {
    System.out.println("");
    pfx="l_";
   }

   String s=pfx+hex_value(i>>2,4)+"  0x"+hex_value(opcode,8)+"   "+disasm(opcode);

   int code_pos=i>>2;

   print_comments(code_pos);

   String ann=get_annotation(code_pos);

   if (ann==null) {
    System.out.println(s);
   } else {
    int len=s.length();
    int pad=ANNOTATION_PADDING-len;

    if (pad>0) s=s+do_pad(pad)+ann;
    System.out.println(s);
   }
  }
 }

 private static int[] load_file(String name) {
  int res[]=null;

  try {
   System.out.println("- loading "+name);

   File f=new File(name);
   int size=(int)f.length();

   res=new int[size];

   FileInputStream in=new FileInputStream(name);

   for(int i=0;i<size;i++) {
    res[i]=in.read();
   }

   in.close();
  } catch(Throwable e) {
   e.printStackTrace();
  }

  return res;
 }

 private static void load_driver() {
  drv=load_file(drv_name);

  String ver=lookup_version();

  if (ver!=null) {
   System.out.println("   ver: "+ver);
  }
 }

 private static void load_files() {
  if (code_name!=null) {
   code=load_file(code_name);
   csize=code.length;
   bblocks=new boolean[csize];
  }

  if (data_name!=null) {
   data=load_file(data_name);
   dsize=data.length;
  }
 }

 private static int read_dword(int tab[],int pos) {
  int res=0;
  res|=(tab[pos++]&0xff)<<24;
  res|=(tab[pos++]&0xff)<<16;
  res|=(tab[pos++]&0xff)<<8;
  res|=(tab[pos++]&0xff);

  return res;
 }

 private static void locate_firmware() {
  try {
   System.out.println("- locating SlimCore firmware");

   int cstart=-1;
   int cend=-1;

   int dstart=-1;
   int dend=-1;

   for(int i=0;i<drv.length-5*4;i+=4) {
    int w1=read_dword(drv,i);
    int w2=read_dword(drv,i+4);

    if ((w1==MAGIC1_CODE_START)&&(w2==MAGIC2_CODE_START)) {
     if (cstart<0) cstart=i+0x04;
    } else

    if ((w1==MAGIC1_CODE_START)&&(w2==0xffffffff)) {
     int w3=read_dword(drv,i+0x10);

     if (w3==MAGIC2_CODE_START) {
      if (cstart<0) cstart=i+0x10;
     }
    } else

    if ((w1==MAGIC1_CODE_END)&&(w2==MAGIC2_CODE_END)) {
     if (dstart<0) dstart=i+0x04;
    }
   }

   if (cstart<0) error("cannot locate firmware code (unsupported STTKDMA version ?)");

   cend=dstart;

   csize=cend-cstart;

   code=new int[csize];
   bblocks=new boolean[csize];

   System.out.println("   code at 0x"+hex_value(cstart,8)+" size "+csize+" (0x"+hex_value(csize,4)+")");
   System.out.println("    - sha1 "+sha1(code));

   for(int i=0;i<csize;i++) {
    code[i]=drv[cstart+i];
   }

   dsize=DATA_SIZE;

   if ((dstart>0)&&(dsize!=0)) {
    data=new int[dsize];

    System.out.println("   data at 0x"+hex_value(dstart,8)+" size "+dsize+" (0x"+hex_value(dsize,4)+")");
    System.out.println("    - sha1 "+sha1(data));
 
    for(int i=0;i<dsize;i++) {
     data[i]=drv[dstart+i];
    }
   }
  } catch(Throwable e) {
   e.printStackTrace();
  }
 }

 private static void save(String name,int buf[]) {
  FileOutputStream fos=null;

  System.out.println("- saving "+name);

  try {
   fos=new FileOutputStream(name,false);
  } catch(Throwable t) {}

  if (fos==null) error("cannot create file");

  try {
   DataOutputStream dos=new DataOutputStream(fos);

   for(int i=0;i<buf.length;i++) {
    dos.write(buf[i]);
   }

   dos.close();
   fos.close();
  } catch(Throwable t) {error("cannot write to file");}
 }

 private static void all_stats() {
  int stat[]=new int[256];

  for(int i=0;i<csize;i+=4) {
   int opcode=opcode(i);
   opcode>>=16;
   stat[opcode]++;
  }

  int num=0;

  for(int i=0;i<stat.length;i++) {
   if (stat[i]!=0) {
    num++;
    System.out.println("opcode 00"+hex_value(i,2)+"xxxx cnt "+stat[i]);
   }
  }

  System.out.println("total "+num+" opcodes");
 }

 private static void unknown_stats() {
  System.out.println("");
  System.out.println("[UNKNOWN INSTRUCTIONS STATS]");

  int stat[]=new int[256];

  for(int i=0;i<csize;i+=4) {
   int opcode=opcode(i);
   String desc=disasm(opcode);

   opcode>>=16;
   
   if (desc.equals("")) stat[opcode]++;
  }

  int num=0;

  for(int i=0;i<stat.length;i++) {
   if (stat[i]!=0) {
    num++;
    System.out.println("opcode 00"+hex_value(i,2)+"xxxx cnt "+stat[i]);
   }
  }

  System.out.println("total "+num+" opcodes");
 }

 public static void usage() {
  System.out.println("usage: SCDisasm [-dis][-m drv|files][-f drv_name][-a ann_name][-c code_file][-d data_file][-stat unk|all][-ext code|data]");
  System.exit(0);
 }

 public static void error(String s) {
  System.out.println("error: "+s);
  System.exit(0);
 }
 
 public static void check_idx(String args[],int idx) {
  if (idx>=args.length) usage();
 }

 public static void add_annotation(int pos,String s) {
  annotations.put(new Integer(pos),s);
 }

 public static String get_annotation(int pos) {
  String s=(String)annotations.get(new Integer(pos));

  return s;
 }

 public static void add_comment(int pos,String s) {
  Vector v=(Vector)comments.get(new Integer(pos));

  if (v==null) {
   v=new Vector();
   comments.put(new Integer(pos),v);
  }

  v.add(s);
 }

 public static Vector get_comments(int pos) {
  Vector v=(Vector)comments.get(new Integer(pos));

  return v;
 }

 public static void init_annotations() {
  try {
   annotations=new Hashtable();
   comments=new Hashtable();

   if (ann_name!=null) {
    System.out.println("- loading annotations "+ann_name);

    FileInputStream fis=new FileInputStream(ann_name);

    BufferedReader b=new BufferedReader(new InputStreamReader(fis,"utf-8"));

    String line=b.readLine();

    int code_pos=0;

    while(line!=null) {
     line=line.trim();
     if (line.startsWith("!")||(line.length()==0)) {
      line=b.readLine();
      continue;
     }

     int idx=line.indexOf(" ");

     if (idx>0) {
      String s_pos=line.substring(0,idx);
      String s_ann=line.substring(idx+1,line.length());

      code_pos=Integer.parseInt(s_pos,16);

      if (s_ann.startsWith(";")) add_annotation(code_pos,s_ann);
       else add_comment(code_pos,s_ann);
     }

     line=b.readLine();
    }
   }
  } catch(Throwable t) {error("parse annotation file");}
 }

 public static void print_comments(int codepos) {
  Vector v=get_comments(codepos);

  if (v!=null) {
   for(int i=0;i<v.size();i++) {
    String s=(String)v.elementAt(i);
    System.out.println(s);
   }
  }
 }

 public static void print_annotation(int codepos) {
  String s=get_annotation(codepos);

  if (s!=null) System.out.println(s);
 }

 public static void prepare() throws Throwable {
  init_syms();
  switch(access_method) {
   case DUMP_FROM_DRIVER:
    load_driver();
    locate_firmware();
    break;
   case DUMP_FROM_FILES:
    load_files();
    break;
  }

  init_annotations();
 }

 public static String disasm(int opcode) {
  Opcode.Decoder d=Opcode.parse(opcode);

  if (d.is_jump()) {
   int target=d.jump_target(opcode);
   mark_target(target);
  }

  String s=d.print(opcode);

  if (d.is_load_or_store()) {
   int reg=d.get_component(opcode,"reg2");
   int val=d.get_component(opcode,"mimm8");

   if ((reg==0)&&(val!=-1)) {
    if ((val&0x80)==0) {
     s+=" // 0x"+hex_value(DSTART+4*val,4);
     if (dsize>0) s+=" = 0x"+hex_value(data(4*val),8);
     if (ann_name==null) {
      String sym=lookup_sym(DSTART+4*val);
      if (sym!=null) s+="  ["+sym+"]";
     }
    } else {
     s+=" // 0x"+hex_value(IOSTART+4*(val&0x7f),4);
     if (ann_name==null) {
      String sym=lookup_sym(IOSTART+4*(val&0x7f));
      if (sym!=null) s+="  ["+sym+"]";
     }
    }
   }
  }

  return s;
 }

 public static void main(String args[]) throws Throwable {
  try {
   System.out.println("/*## (c) SECURITY EXPLORATIONS    2011 poland                                #*/");
   System.out.println("/*##     http://www.security-explorations.com                                #*/");
   System.out.println("SlimCore disassembler");

   int cmd=0;
   access_method=0;

   int idx=0;

   if (args.length==0) usage();

   while(idx<args.length) {
    String arg=args[idx++];

    if (arg.equals("-dis")) {
     cmd=DISASM;
    } else

    if (arg.equals("-ext")) {
     check_idx(args,idx);
     String opt=args[idx++];

     if (opt.equals("code")) cmd=EXTRACT_CODE;
      else
     if (opt.equals("data")) cmd=EXTRACT_DATA;
      else usage();
    } else 

    if (arg.equals("-stat")) {
     check_idx(args,idx);
     String opt=args[idx++];

     if (opt.equals("unk")) cmd=STATS_PRINT_UNK;
      else
     if (opt.equals("all")) cmd=STATS_PRINT_ALL;
      else usage();
    } else

    if (arg.equals("-f")) {
     check_idx(args,idx);
     drv_name=args[idx++];
    } else

    if (arg.equals("-a")) {
     check_idx(args,idx);
     ann_name=args[idx++];
    } else

    if (arg.equals("-c")) {
     check_idx(args,idx);
     code_name=args[idx++];
    } else

    if (arg.equals("-d")) {
     check_idx(args,idx);
     data_name=args[idx++];
    } else

    if (arg.equals("-m")) {
     check_idx(args,idx);
     String opt=args[idx++];

     if (opt.equals("drv")) access_method=DUMP_FROM_DRIVER;
      else
     if (opt.equals("files")) access_method=DUMP_FROM_FILES;
      else usage();
    } else usage();
   }

   if (cmd==0)  error("no command");

   if (drv_name==null) drv_name=STTKDMA_DRIVER;

   fulldisasm=true;

   prepare();

   switch(cmd) {
    case DISASM:
     dump_code();
     break;
    case EXTRACT_CODE:
     if (code_name!=null) {
      save(code_name,code);
     }
     break;
    case EXTRACT_DATA:
     if (data_name!=null) {
      save(data_name,data);
     }
     break;
    case STATS_PRINT_UNK:
     unknown_stats();
     break;
    case STATS_PRINT_ALL:
     all_stats();
     break;
   }
  } catch(Throwable e) {
   e.printStackTrace();
  } 
 }
}
